package at.ac.tuwien.sepm.assignment.individual.integration;

import static org.junit.jupiter.api.Assertions.*;

import at.ac.tuwien.sepm.assignment.individual.endpoint.dto.HorseDto;
import at.ac.tuwien.sepm.assignment.individual.enums.ERace;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.web.server.LocalServerPort;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.core.io.ClassPathResource;
import org.springframework.http.*;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import java.io.File;
import java.sql.Date;
import java.util.List;

@SpringBootTest(webEnvironment = SpringBootTest.WebEnvironment.RANDOM_PORT)
@ActiveProfiles("test")
public class HorseEndpointTest {
    @Value("${spring.upload.path}")
    private String FILE_BASE_PATH;
    private static final RestTemplate REST_TEMPLATE = new RestTemplate();
    private static final String BASE_URL = "http://localhost:";
    private static final String HORSE_URL = "/horses";

    @LocalServerPort
    private int port;

    @Test
    @DisplayName("Searching for all horses with no filters, given two horses should return HTTP 200 and both horses")
    public void searchingHorses_noFiltersGivenTwoHorses_shouldReturnStatus200AndHorses() {
        // Create the horses
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", Long.valueOf(0));

        HttpEntity<HorseDto> request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> firstHorse = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        newHorse = new HorseDto("Katrina", "Not a fast enough horse", (short) 3, Date.valueOf("2005-01-01"), ERace.PAINT, "files/example.png", Long.valueOf(0));

        request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> secondHorse = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        // Get all horses and save in list
        // https://stackoverflow.com/a/31947188
        ResponseEntity<List<HorseDto>> allHorses = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.GET, null, new ParameterizedTypeReference<List<HorseDto>>() {});

        assertEquals(allHorses.getStatusCode(), HttpStatus.OK);
        assertTrue(allHorses.getBody().contains(firstHorse.getBody()));
        assertTrue(allHorses.getBody().contains(secondHorse.getBody()));
    }

    @Test
    @DisplayName("Searching for all horses with correct filters, given two horses should return HTTP 200 and one horse")
    public void searchingHorses_withCorrectFiltersGivenTwoHorses_shouldReturnStatus200AndHorse() {
        // Create the horses
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", Long.valueOf(0));

        HttpEntity<HorseDto> request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> firstHorse = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        newHorse = new HorseDto("Katrina", "Not a fast enough horse", (short) 3, Date.valueOf("2005-01-01"), ERace.PAINT, "files/example.png", Long.valueOf(0));

        request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> secondHorse = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        // Get all horses and save in list
        // https://stackoverflow.com/a/25434451
        UriComponentsBuilder builder = UriComponentsBuilder.fromHttpUrl(BASE_URL + port + HORSE_URL)
            .queryParam("name", "Zephyr")
            .queryParam("description", "Nice")
            .queryParam("score", "4")
            .queryParam("birthday", "2020-03-23")
            .queryParam("race", ERace.APPALOOSA.name());

        ResponseEntity<List<HorseDto>> allHorses = REST_TEMPLATE.exchange(
            builder.toUriString(),
            HttpMethod.GET,
            null,
            new ParameterizedTypeReference<List<HorseDto>>() {}
        );

        assertEquals(allHorses.getStatusCode(), HttpStatus.OK);
        assertTrue(allHorses.getBody().contains(firstHorse.getBody()));
    }

    @Test
    @DisplayName("Searching for all horses with incorrect filters, given two horses should return HTTP 400")
    public void searchingHorses_withIncorrectFiltersGivenTwoHorses_shouldReturnStatus400() {
        // Create the horses
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", Long.valueOf(0));

        HttpEntity<HorseDto> request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> firstHorse = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        newHorse = new HorseDto("Katrina", "Not a fast enough horse", (short) 3, Date.valueOf("2005-01-01"), ERace.PAINT, "files/example.png", Long.valueOf(0));

        request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> secondHorse = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        // Get all horses and save in list
        // https://stackoverflow.com/a/25434451
        UriComponentsBuilder builder = UriComponentsBuilder.fromHttpUrl(BASE_URL + port + HORSE_URL)
            .queryParam("name", "Zephyr")
            .queryParam("description", "Nice")
            .queryParam("score", "10")
            .queryParam("birthday", "test")
            .queryParam("race", "test");

        assertThrows(HttpClientErrorException.BadRequest.class, () ->
            REST_TEMPLATE
                .exchange(builder.toUriString(), HttpMethod.GET, null, new ParameterizedTypeReference<HorseDto>() {}));
    }

    @Test
    @DisplayName("Adding a new horse with the correct parameters will return HTTP 201 and the new HorseDto")
    public void addingNewHorse_correctParameters_shouldReturnStatus201AndHorse() {
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);

        HttpEntity<HorseDto> request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        // Compare everything except ids and timestamps
        assertEquals(response.getStatusCode(), HttpStatus.CREATED);
        assertEquals(newHorse.getName(), response.getBody().getName());
        assertEquals(newHorse.getDescription(), response.getBody().getDescription());
        assertEquals(newHorse.getScore(), response.getBody().getScore());
        assertEquals(newHorse.getBirthday().toString(), response.getBody().getBirthday().toString());
        assertEquals(newHorse.getImagePath(), response.getBody().getImagePath());
        assertEquals(newHorse.getOwner(), response.getBody().getOwner());
    }

    @Test
    @DisplayName("Updating a horse with the correct parameters will return 200 and the new HorseDto")
    public void updatingNewHorse_correctParameters_shouldReturnStatus200AndHorse() {
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);

        // Create a new horse
        HttpEntity<HorseDto> request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        // Update the horse
        newHorse.setId(response.getBody().getId());
        newHorse.setName("Katrina");
        newHorse.setDescription("Fast horse");
        newHorse.setScore((short) 3);
        newHorse.setBirthday(Date.valueOf("2005-01-01"));
        newHorse.setRace(ERace.MORGAN);
        newHorse.setImagePath("files/katrina.png");
        newHorse.setOwner(null);

        request = new HttpEntity<>(newHorse);
        response = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL + '/' + newHorse.getId(), HttpMethod.PUT, request, HorseDto.class);

        // Compare everything except timestamps
        assertEquals(response.getStatusCode(), HttpStatus.OK);
        assertEquals(newHorse.getId(), response.getBody().getId());
        assertEquals(newHorse.getName(), response.getBody().getName());
        assertEquals(newHorse.getDescription(), response.getBody().getDescription());
        assertEquals(newHorse.getScore(), response.getBody().getScore());
        assertEquals(newHorse.getBirthday().toString(), response.getBody().getBirthday().toString());
        assertEquals(newHorse.getImagePath(), response.getBody().getImagePath());
        assertEquals(newHorse.getOwner(), response.getBody().getOwner());
    }

    @Test
    @DisplayName("Uploading an image in the correct format will return HTTP 201")
    public void addingNewImage_correctFormat_shouldReturnStatus201() {
        try {
            // More information:
            // https://www.baeldung.com/spring-rest-template-multipart-upload
            // https://github.com/spring-guides/gs-uploading-files/blob/master/complete/src/test/java/com/example/uploadingfiles/FileUploadIntegrationTests.java

            // Get the file
            ClassPathResource image = new ClassPathResource("horse.jpg", getClass());

            // Set the body
            MultiValueMap<String, Object> body = new LinkedMultiValueMap<>();
            body.add("file", image);

            // Create the HTTP Entity
            HttpEntity<MultiValueMap<String, Object>> request = new HttpEntity<>(body);

            ResponseEntity<String> response = REST_TEMPLATE.postForEntity(BASE_URL + port + HORSE_URL + "/upload", request, String.class);
            assertEquals(HttpStatus.CREATED, response.getStatusCode());
        } finally {
            new File(FILE_BASE_PATH + "horse.jpg").delete();
        }
    }

    @Test
    @DisplayName("Deleting an existing horse will return HTTP 204")
    public void deletingHorse_existing_shouldReturnStatus204() {
        // Create the horse
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);

        HttpEntity<HorseDto> request = new HttpEntity<>(newHorse);
        ResponseEntity<HorseDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        // Delete and test if deleted
        ResponseEntity res = REST_TEMPLATE
            .exchange(BASE_URL + port + HORSE_URL + '/' + response.getBody().getId(), HttpMethod.DELETE, null, new ParameterizedTypeReference<HorseDto>() {});

        assertEquals(res.getStatusCode(), HttpStatus.NO_CONTENT);
    }

    @Test
    @DisplayName("Deleting an nonexistent horse will return HTTP 404")
    public void deletingHorse_nonexistent_shouldReturnStatus404() {
        assertThrows(HttpClientErrorException.NotFound.class, () ->
            REST_TEMPLATE
                .exchange(BASE_URL + port + HORSE_URL + "/0", HttpMethod.DELETE, null, new ParameterizedTypeReference<HorseDto>() {}));
    }
}

