package at.ac.tuwien.sepm.assignment.individual.integration;

import static at.ac.tuwien.sepm.assignment.individual.base.TestData.HORSE_URL;
import static org.junit.jupiter.api.Assertions.*;

import at.ac.tuwien.sepm.assignment.individual.endpoint.dto.HorseDto;
import at.ac.tuwien.sepm.assignment.individual.endpoint.dto.OwnerDto;
import at.ac.tuwien.sepm.assignment.individual.enums.ERace;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.web.server.LocalServerPort;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.*;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import java.sql.Date;
import java.util.List;

@SpringBootTest(webEnvironment = SpringBootTest.WebEnvironment.RANDOM_PORT)
@ActiveProfiles("test")
public class OwnerEndpointTest {
    @Value("${spring.upload.path}")
    private String FILE_BASE_PATH;
    private static final RestTemplate REST_TEMPLATE = new RestTemplate();
    private static final String BASE_URL = "http://localhost:";
    private static final String OWNER_URL = "/owners";

    @LocalServerPort
    private int port;


    @Test
    @DisplayName("Searching for all owners with no filters, given two owners should return HTTP 200 and both ownerss")
    public void searchingOwners_noFiltersGivenTwoOwners_shouldReturnStatus200AndOwners() {
        // Create the owners
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> firstOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        newOwner = new OwnerDto("Gigachad");

        request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> secondOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Get all owners and save in list
        // https://stackoverflow.com/a/31947188
        ResponseEntity<List<OwnerDto>> allOwners = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.GET, null, new ParameterizedTypeReference<List<OwnerDto>>() {});

        assertEquals(allOwners.getStatusCode(), HttpStatus.OK);
        assertTrue(allOwners.getBody().contains(firstOwner.getBody()));
        assertTrue(allOwners.getBody().contains(secondOwner.getBody()));
    }

    @Test
    @DisplayName("Searching for all owners with correct filters, given two owners should return HTTP 200 and one owner")
    public void searchingOwners_withCorrectFiltersGivenTwoOwners_shouldReturnStatus200AndOwner() {
        // Create the owners
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> firstOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        newOwner = new OwnerDto("Lad");

        request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> secondOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Get all owners and save in list
        // https://stackoverflow.com/a/25434451
        UriComponentsBuilder builder = UriComponentsBuilder.fromHttpUrl(BASE_URL + port + OWNER_URL)
            .queryParam("name", "chad");

        ResponseEntity<List<OwnerDto>> allOwners = REST_TEMPLATE.exchange(
            builder.toUriString(),
            HttpMethod.GET,
            null,
            new ParameterizedTypeReference<List<OwnerDto>>() {}
        );

        assertEquals(allOwners.getStatusCode(), HttpStatus.OK);
        assertTrue(allOwners.getBody().contains(firstOwner.getBody()));
    }

    @Test
    @DisplayName("Searching for all owners with incorrect filters, given two owners should return HTTP 400")
    public void searchingOwners_withIncorrectFiltersGivenTwoOwners_shouldReturnStatus400() {
        // Create the owners
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> firstOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        newOwner = new OwnerDto("Lad");

        request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> secondOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Get all owners and save in list
        // https://stackoverflow.com/a/25434451
        UriComponentsBuilder builder = UriComponentsBuilder.fromHttpUrl(BASE_URL + port + OWNER_URL)
            .queryParam("name", "");

        assertThrows(HttpClientErrorException.BadRequest.class, () ->
            REST_TEMPLATE
                .exchange(builder.toUriString(), HttpMethod.GET, null, new ParameterizedTypeReference<OwnerDto>() {}));
    }

    @Test
    @DisplayName("Getting horses of an owner with one horse should return HTTP 200 and the horse")
    public void gettingHorsesOfOwner_oneHorseAssigned_shouldReturnStatus200AndHorse() {
        // Create the owner
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        OwnerDto savedOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class).getBody();

        // Create the horse
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", savedOwner.getId());

        request = new HttpEntity(newHorse);
        ResponseEntity<HorseDto> savedHorse = REST_TEMPLATE.exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        ResponseEntity<List<HorseDto>> allHorses = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL + "/" + savedOwner.getId() + "/horses", HttpMethod.GET, null, new ParameterizedTypeReference<List<HorseDto>>() {});

        // Temporary, until db mapping is fixed
        savedHorse.getBody().setOwner(null);

        assertEquals(allHorses.getStatusCode(), HttpStatus.OK);
        assertTrue(allHorses.getBody().contains(savedHorse.getBody()));
    }

    @Test
    @DisplayName("Getting horses of an owner without horses should return HTTP 404")
    public void gettingHorsesOfOwner_noHorsesAssigned_shouldReturnStatus404() {
        // Create the owner
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        OwnerDto savedOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class).getBody();

        assertThrows(HttpClientErrorException.NotFound.class, () -> REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL + "/" + savedOwner.getId() + "/horses", HttpMethod.GET, null, new ParameterizedTypeReference<List<HorseDto>>() {}));
    }

    @Test
    @DisplayName("Adding a new owner with the correct parameters will return HTTP 201 and the new OwnerDto")
    public void addingNewOwner_correctParameters_shouldReturnStatus201AndOwner() {
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Compare everything except ids and timestamps
        assertEquals(response.getStatusCode(), HttpStatus.CREATED);
        assertEquals(newOwner.getName(), response.getBody().getName());
    }

    @Test
    @DisplayName("Updating a owner with the correct parameters will return 200 and the new OwnerDto")
    public void updatingNewOwner_correctParameters_shouldReturnStatus200AndOwner() {
        OwnerDto newOwner = new OwnerDto("Chad");

        // Create a new owner
        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Update the owner
        newOwner.setId(response.getBody().getId());
        newOwner.setName("Gigachad");

        request = new HttpEntity<>(newOwner);
        response = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL + '/' + newOwner.getId(), HttpMethod.PUT, request, OwnerDto.class);

        // Compare everything except timestamps
        assertEquals(response.getStatusCode(), HttpStatus.OK);
        assertEquals(newOwner.getId(), response.getBody().getId());
        assertEquals(newOwner.getName(), response.getBody().getName());
    }

    @Test
    @DisplayName("Deleting an existing owner without horses will return HTTP 204")
    public void deletingOwner_existingNoOwnersOwned_shouldReturnStatus204() {
        // Create the owner
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Delete and test if deleted
        ResponseEntity res = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL + '/' + response.getBody().getId(), HttpMethod.DELETE, null, new ParameterizedTypeReference<OwnerDto>() {});

        assertEquals(res.getStatusCode(), HttpStatus.NO_CONTENT);
    }

    @Test
    @DisplayName("Deleting an existing owner without horses will return HTTP 204")
    public void deletingOwner_existingNoHorsesOwned_shouldReturnStatus204() {
        // Create the owner
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        ResponseEntity<OwnerDto> response = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class);

        // Delete and test if deleted
        ResponseEntity res = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL + '/' + response.getBody().getId(), HttpMethod.DELETE, null, new ParameterizedTypeReference<OwnerDto>() {});

        assertEquals(res.getStatusCode(), HttpStatus.NO_CONTENT);
    }

    @Test
    @DisplayName("Deleting an nonexistent owner will return HTTP 404")
    public void deletingOwner_nonexistent_shouldReturnStatus404() {
        assertThrows(HttpClientErrorException.NotFound.class, () ->
            REST_TEMPLATE
                .exchange(BASE_URL + port + OWNER_URL + "/0", HttpMethod.DELETE, null, new ParameterizedTypeReference<OwnerDto>() {}));
    }

    @Test
    @DisplayName("Deleting an exiting owner with horses will return HTTP 403")
    public void deletingOwner_existingHorsesOwned_shouldReturnStatus403() {
        // Create the owner
        OwnerDto newOwner = new OwnerDto("Chad");

        HttpEntity<OwnerDto> request = new HttpEntity<>(newOwner);
        OwnerDto savedOwner = REST_TEMPLATE
            .exchange(BASE_URL + port + OWNER_URL, HttpMethod.POST, request, OwnerDto.class).getBody();

        // Create the horse
        HorseDto newHorse = new HorseDto("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", savedOwner.getId());

        request = new HttpEntity(newHorse);
        REST_TEMPLATE.exchange(BASE_URL + port + HORSE_URL, HttpMethod.POST, request, HorseDto.class);

        assertThrows(HttpClientErrorException.Forbidden.class, () ->
            REST_TEMPLATE
                .exchange(BASE_URL + port + OWNER_URL + "/" + savedOwner.getId(), HttpMethod.DELETE, null, new ParameterizedTypeReference<OwnerDto>() {}));
    }
}
