package at.ac.tuwien.sepm.assignment.individual.unit.persistence;

import static org.junit.jupiter.api.Assertions.*;

import at.ac.tuwien.sepm.assignment.individual.entity.Horse;
import at.ac.tuwien.sepm.assignment.individual.enums.ERace;
import at.ac.tuwien.sepm.assignment.individual.exception.NotFoundException;
import at.ac.tuwien.sepm.assignment.individual.persistence.HorseDao;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;

import java.io.IOException;
import java.sql.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public abstract class HorseDaoTestBase {

    @Autowired
    HorseDao horseDao;

    @Test
    @DisplayName("Getting all horses given two horses should return a list with the horses")
    public void gettingAllHorses_givenTwoHorses_shouldReturnHorses() {
        // Create the horses
        Horse firstHorse = horseDao.addHorse(new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", Long.valueOf(0)));
        Horse secondHorse = horseDao.addHorse(new Horse("Katrina", "Bad horse", (short) 1, Date.valueOf("2005-01-01"), ERace.APPALOOSA, "files/example.png", Long.valueOf(0)));

        // Test if the horses are present
        List<Horse> allHorses = horseDao.getAll();
        assertTrue(allHorses.contains(firstHorse));
        assertTrue(allHorses.contains(secondHorse));
    }

    @Test
    @DisplayName("Searching all horses with correct filters given two horses should return a list with the horse")
    public void searchingHorses_withCorrectFiltersGivenTwoHorses_shouldReturnHorse() {
        // Create the horses
        Horse firstHorse = horseDao.addHorse(new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", Long.valueOf(0)));
        Horse secondHorse = horseDao.addHorse(new Horse("Katrina", "Bad horse", (short) 1, Date.valueOf("2005-01-01"), ERace.APPALOOSA, "files/example.png", Long.valueOf(0)));

        // Test if the horses are present
        Map<String, String> filters = new HashMap<String, String>();
        filters.put("name", "Zephyr");
        filters.put("description", "Nice");
        filters.put("score", "4");
        filters.put("birthday", "2020-03-23");
        filters.put("race", ERace.APPALOOSA.name());

        List<Horse> allHorses = horseDao.getFiltered(filters);
        assertTrue(allHorses.contains(firstHorse));
    }

    @Test
    @DisplayName("Searching all horses with incorrect filters given two horses should throw NotFoundException")
    public void searchingHorses_withIncorrectFiltersGivenTwoHorses_shouldThrowNotFound() {
        // Create the horses
        Horse firstHorse = horseDao.addHorse(new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", Long.valueOf(0)));
        Horse secondHorse = horseDao.addHorse(new Horse("Katrina", "Bad horse", (short) 1, Date.valueOf("2005-01-01"), ERace.APPALOOSA, "files/example.png", Long.valueOf(0)));

        // Test if the horses are present
        Map<String, String> filters = new HashMap<String, String>();
        filters.put("name", "Tester Horse");
        filters.put("description", "A horse for testing");
        filters.put("score", "1");
        filters.put("birthday", "2020-02-01");
        filters.put("race", ERace.APPALOOSA.name());

        assertThrows(NotFoundException.class, () -> horseDao.getFiltered(filters));
    }

    @Test
    @DisplayName("Adding a new horse with the correct parameters should return the horse")
    public void addingNewHorse_correctParameters_shouldReturnHorse() {
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);
        Horse savedHorse = horseDao.addHorse(newHorse);
        assertEquals(newHorse, savedHorse);
    }

    @Test
    @DisplayName("Adding a new horse with the incorrect parameters should throw DataAccessException")
    public void addingNewHorse_incorrectParameters_shouldThrowDataAccess() {
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 80, Date.valueOf("2020-01-01"), null, null, null);
        assertThrows(DataAccessException.class, () -> horseDao.addHorse(newHorse));
    }

    @Test
    @DisplayName("Updating a horse with the correct parameters should return the horse")
    public void updatingHorse_correctParameters_shouldReturnHorse() throws IOException {
        // Create horse
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);
        Horse savedHorse = horseDao.addHorse(newHorse);

        // Update horse
        newHorse.setId(savedHorse.getId());
        newHorse.setName("Katrina");
        newHorse.setDescription("Fast horse");
        newHorse.setScore((short) 3);
        newHorse.setBirthday(Date.valueOf("2005-01-01"));
        newHorse.setRace(ERace.MORGAN);
        newHorse.setImagePath("files/katrina.png");
        newHorse.setOwner(null);
        Horse updatedHorse = horseDao.updateHorse(newHorse);

        // Compare everything except updated timestamp
        assertEquals(updatedHorse.getId(), newHorse.getId());
        assertEquals(updatedHorse.getName(), newHorse.getName());
        assertEquals(updatedHorse.getDescription(), newHorse.getDescription());
        assertEquals(updatedHorse.getScore(), newHorse.getScore());
        assertEquals(updatedHorse.getBirthday().toString(), newHorse.getBirthday().toString());
        assertEquals(updatedHorse.getImagePath(), newHorse.getImagePath());
        assertEquals(updatedHorse.getOwner(), newHorse.getOwner());
        assertEquals(updatedHorse.getCreatedAt(), newHorse.getCreatedAt());
    }

    @Test
    @DisplayName("Updating a horse with the incorrect parameters should throw DataAccessException")
    public void updatingHorse_incorrectParameters_shouldThrowDataAccess() {
        // Create horse
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);
        Horse savedHorse = horseDao.addHorse(newHorse);

        // Update horse
        newHorse.setId(savedHorse.getId());
        newHorse.setName("Katrina");
        newHorse.setDescription("Fast horse");
        newHorse.setScore((short) 80);
        newHorse.setBirthday(Date.valueOf("2005-01-01"));
        newHorse.setRace(null);
        newHorse.setImagePath(null);
        newHorse.setOwner(null);
        assertThrows(DataAccessException.class, () -> horseDao.updateHorse(newHorse));
    }

    @Test
    @DisplayName("Deleting an existing horse should delete the horse")
    public void deletingHorse_existing_shouldDeleteHorse() throws IOException {
        // Create the horse
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", null);
        Horse savedHorse = horseDao.addHorse(newHorse);

        // Delete the horse
        horseDao.deleteHorse(savedHorse.getId());

        // Check if deleted
        assertThrows(NotFoundException.class, () -> horseDao.findOneById(savedHorse.getId()));
    }

    @Test
    @DisplayName("Deleting an nonexistent horse should throw NotFoundException")
    public void deletingHorse_nonexistent_shouldThrowNotFound() throws IOException {
        assertThrows(NotFoundException.class, () -> horseDao.deleteHorse(null));
    }
}
