package at.ac.tuwien.sepm.assignment.individual.unit.persistence;

import static org.junit.jupiter.api.Assertions.*;

import at.ac.tuwien.sepm.assignment.individual.entity.Horse;
import at.ac.tuwien.sepm.assignment.individual.entity.Owner;
import at.ac.tuwien.sepm.assignment.individual.enums.ERace;
import at.ac.tuwien.sepm.assignment.individual.exception.NotFoundException;
import at.ac.tuwien.sepm.assignment.individual.persistence.HorseDao;
import at.ac.tuwien.sepm.assignment.individual.persistence.OwnerDao;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.DataIntegrityViolationException;

import java.io.IOException;
import java.sql.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public abstract class OwnerDaoTestBase {

    @Autowired
    OwnerDao ownerDao;

    @Autowired
    HorseDao horseDao;

    @Test
    @DisplayName("Getting all owners given two owners should return a list with the owners")
    public void gettingAllOwners_givenTwoOwners_shouldReturnOwners() {
        // Create the owners
        Owner firstOwner = ownerDao.addOwner(new Owner("Chad"));
        Owner secondOwner = ownerDao.addOwner(new Owner("Lad"));

        // Test if the owners are present
        List<Owner> allOwners = ownerDao.getAll();
        assertTrue(allOwners.contains(firstOwner));
        assertTrue(allOwners.contains(secondOwner));
    }

    @Test
    @DisplayName("Searching all owners with correct filters given two owners should return a list with the owner")
    public void searchingOwners_withCorrectFiltersGivenTwoOwners_shouldReturnOwner() {
        // Create the owners
        Owner firstOwner = ownerDao.addOwner(new Owner("Brad"));
        Owner secondOwner = ownerDao.addOwner(new Owner("Becky"));

        // Test if the owners are present
        Map<String, String> filters = new HashMap<String, String>();
        filters.put("name", "brad");

        List<Owner> allOwners = ownerDao.getFiltered(filters);
        assertTrue(allOwners.contains(firstOwner));
    }

    @Test
    @DisplayName("Searching all owners with incorrect filters given two owners should throw NotFoundException")
    public void searchingOwners_withIncorrectFiltersGivenTwoOwners_shouldThrowNotFound() {
        // Create the owners
        Owner firstOwner = ownerDao.addOwner(new Owner("Brad"));
        Owner secondOwner = ownerDao.addOwner(new Owner("Stacy"));

        // Test if the owners are present
        Map<String, String> filters = new HashMap<String, String>();
        filters.put("name", "Tester Owner");

        assertThrows(NotFoundException.class, () -> ownerDao.getFiltered(filters));
    }

    @Test
    @DisplayName("Finding owner by non-existing ID should throw NotFoundException")
    public void findingOwnerById_nonExisting_shouldThrowNotFoundException() {
        assertThrows(NotFoundException.class,
            () -> ownerDao.findOneById(0L));
    }

    @Test
    @DisplayName("Getting horses of owner with one horse should return the Horse")
    public void gettingHorsesOfOwner_oneHorseAssigned_shouldReturnHorse() {
        Owner savedOwner = ownerDao.addOwner(new Owner("Brad"));

        // Create the horse
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", savedOwner.getId());
        Horse savedHorse = horseDao.addHorse(newHorse);

        // Temporary, until the database mapping is fixed
        savedHorse.setOwner(null);
        assertTrue(ownerDao.getOwnedHorses(savedOwner.getId()).contains(newHorse));
    }

    @Test
    @DisplayName("Getting horses of owner without horses should throw NotFoundException")
    public void gettingHorsesOfOwner_noHorsesAssigned_shouldThrowNotFound() {
        Owner savedOwner = ownerDao.addOwner(new Owner("Brad"));

        assertThrows(NotFoundException.class, () -> ownerDao.getOwnedHorses(savedOwner.getId()));
    }

    @Test
    @DisplayName("Adding a new owner with the correct parameters should return the owner")
    public void addingNewOwner_correctParameters_shouldReturnOwner() {
        Owner newOwner = new Owner("Chad");
        Owner savedOwner = ownerDao.addOwner(newOwner);
        assertEquals(newOwner, savedOwner);
    }

    @Test
    @DisplayName("Adding a new owner with the incorrect parameters should throw DataAccessException")
    public void addingNewOwner_incorrectParameters_shouldThrowDataAccess() {
        Owner newOwner = new Owner("");
        assertThrows(DataAccessException.class, () -> ownerDao.addOwner(newOwner));
    }

    @Test
    @DisplayName("Updating a owner with the correct parameters should return the owner")
    public void updatingOwner_correctParameters_shouldReturnOwner() throws IOException {
        // Create owner
        Owner newOwner = new Owner("Chad");
        Owner savedOwner = ownerDao.addOwner(newOwner);

        // Update owner
        newOwner.setId(savedOwner.getId());
        newOwner.setName("Gigachad");
        Owner updatedOwner = ownerDao.updateOwner(newOwner);

        // Compare everything except updated timestamp
        assertEquals(updatedOwner.getId(), newOwner.getId());
        assertEquals(updatedOwner.getName(), newOwner.getName());
        assertEquals(updatedOwner.getCreatedAt(), newOwner.getCreatedAt());
    }

    @Test
    @DisplayName("Updating a owner with the incorrect parameters should throw DataAccessException")
    public void updatingOwner_incorrectParameters_shouldThrowDataAccess() {
        // Create owner
        Owner newOwner = new Owner("Chad");
        Owner savedOwner = ownerDao.addOwner(newOwner);

        // Update owner
        newOwner.setId(savedOwner.getId());
        newOwner.setName("");
        assertThrows(DataAccessException.class, () -> ownerDao.updateOwner(newOwner));
    }

    @Test
    @DisplayName("Deleting an existing owner without horses should delete the owner")
    public void deletingOwner_existingOwnerNoHorsesOwned_shouldDeleteOwner() throws IOException {
        // Create the owner
        Owner newOwner = new Owner("Chad");
        Owner savedOwner = ownerDao.addOwner(newOwner);

        // Delete the owner
        ownerDao.deleteOwner(savedOwner.getId());

        // Check if deleted
        assertThrows(NotFoundException.class, () -> ownerDao.findOneById(savedOwner.getId()));
    }

    @Test
    @DisplayName("Deleting an nonexistent owner should throw NotFoundException")
    public void deletingOwner_nonexistent_shouldThrowNotFound() throws IOException {
        assertThrows(NotFoundException.class, () -> ownerDao.deleteOwner(null));
    }

    @Test
    @DisplayName("Deleting an existing owner with horses should throw DataIntegrityViolationException")
    public void deletingHorse_existing_shouldDeleteHorse() {
        // Create the owner
        Owner newOwner = new Owner("Chad");
        Owner savedOwner = ownerDao.addOwner(newOwner);

        // Create the horse
        Horse newHorse = new Horse("Zephyr", "Nice horse", (short) 4, Date.valueOf("2020-01-01"), ERace.APPALOOSA, "files/test.png", savedOwner.getId());
        Horse savedHorse = horseDao.addHorse(newHorse);

        // Delete the owner
        assertThrows(DataIntegrityViolationException.class, () -> ownerDao.deleteOwner(savedOwner.getId()));
    }
}
