import { Component, OnInit } from '@angular/core';
import { Horse } from '../../dto/horse';
import { HorseService } from '../../service/horse.service';
import { HttpParams } from '@angular/common/http';

@Component({
  selector: 'app-list-horses',
  templateUrl: './list-horses.component.html',
  styleUrls: ['./list-horses.component.scss']
})
export class ListHorsesComponent implements OnInit {
  error = false;
  errorMessage = '';
  horses: Array<Horse>;
  filters: {
    name: string,
    description: string,
    birthday: string,
    race: string,
    score: string
  };

  constructor(private horseService: HorseService) { }

  ngOnInit(): void {
    // Nullify queries
    this.filters = {
      name: "",
      description: "",
      birthday: null,
      race: "",
      score: "",
    };
    this.loadAllHorses();
  }

  /**
   * Error flag will be deactivated, which clears the error message
   */
  vanishError() {
    this.error = false;
  }

  /**
   * Loads all horses from the database
   */
  private loadAllHorses() {
    console.log("GET all horses from API");
    this.horseService.getAllHorses().subscribe(
      (horses: Array<Horse>) => {
        this.horses = horses;
      },
      error => {
        this.defaultServiceErrorHandling(error);
      }
    );
  }

  /**
   * Load filtered horses from the database
   */
  public loadFilteredHorses() {
    console.log("GET filtered horses from API");

    console.log(this.filters);
    // Create HttpParams for the sorting
    let params: HttpParams = new HttpParams();
    if(this.filters.name != null && this.filters.name.length !== 0)
      params = params.set('name', this.filters.name);

    if(this.filters.description != null && this.filters.description.length !== 0)
      params = params.set('description', this.filters.description);

    if(this.filters.score != null && this.filters.score.length !== 0)
      params = params.set('score', this.filters.score);

    if(this.filters.birthday !== null)
      params = params.set('birthday', this.filters.birthday);

    if(this.filters.race !== null && this.filters.race.length !== 0)
      params = params.set('race', this.filters.race);

    this.horseService.getFilteredHorses(params).subscribe(
      (horses: Array<Horse>) => {
        this.horses = horses;
      },
      error => {
        this.horses = null;
        this.defaultServiceErrorHandling(error);
      }
    );
  }

  public deleteHorse(id: number, name: string) {
    if(confirm("Are you sure you want to delete " + name + ". This action is irreversible.")) {
      this.horseService.deleteHorse(id).subscribe(
        (res) => {
          console.log('Successfully deleted horse');
          this.loadAllHorses();
        }
      );
    }
  }

  private defaultServiceErrorHandling(error: any) {
    console.log(error);
    this.error = true;
    if (error.status === 0) {
      // If status is 0, the backend is probably down
      this.errorMessage = 'The backend seems not to be reachable';
    } else if (error.error.message === 'No message available') {
      // If no detailed error message is provided, fall back to the simple error name
      this.errorMessage = error.error.error;
    } else {
      this.errorMessage = error.error.message;
    }
  }
}
