import { Component, OnInit } from '@angular/core';
import { ActivatedRoute, Params} from '@angular/router';
import { Horse } from '../../dto/horse';
import { Owner } from '../../dto/owner';
import { HorseService } from '../../service/horse.service';
import { OwnerService } from '../../service/owner.service';

@Component({
  selector: 'app-update-horse',
  templateUrl: './update-horse.component.html',
  styleUrls: ['./update-horse.component.scss']
})
export class UpdateHorseComponent implements OnInit {
  error = false;
  success: string;
  errorMessage = '';
  horse: Horse = new Horse(null, null, null, null, null, null, null, null);
  owners: Array<Owner>;
  imageToUpload: File = null;

  constructor(private horseService: HorseService, private route: ActivatedRoute, private ownerService: OwnerService) { }

  ngOnInit(): void {
    // Extract id from url
    const horseId: string = this.route.snapshot.paramMap.get('id');

    // Get current value of the horse and save it
    this.horseService.getHorseById(parseInt(horseId, 10)).subscribe(
      (horse: Horse) => {
        this.horse = horse;
      },
      error => {
        this.defaultServiceErrorHandling(error);
      }
    );

    // Get all owners and save them
    this.ownerService.getAllOwners().subscribe(
      (owners: Array<Owner>) => {
        this.owners = owners;
      },
      error => {
        this.owners = null;
      }
    );
  }

  /**
   * Will be called on a click on the success alert close button
   */
  public vanishAlert() {
    this.success = null;
  }

  /**
   * Will be called on a click on the error alert close button
   */
  public vanishError() {
    this.errorMessage = null;
  }

  /**
   * Handles the input on the image input.
   * Inspired from:
   * https://stackoverflow.com/questions/47936183/angular-file-upload
   * @param files
   */
  public handleImageInput(files: FileList) {
    var imageFile: File = files.item(0);

    // Generate a random bytes name to avoid conflicts
    var fileExt: string = imageFile.name.split('.').pop();
    var newFileName: string = this.generateHex(20) + '.' + fileExt;
    this.horse.imagePath = newFileName;

    // Upload the file
    this.uploadFile(imageFile, newFileName)
  }

  /**
   * Updates a horse and loads it
   * @param horse
   */
  public updateHorse() {
    console.log("PUT horse to API")
    console.log(this.horse);
    // TODO: Make it possible for the horse to be added with an owner
    this.horseService.updateHorse(this.horse).subscribe(
      (result: Horse) => {
        this.success = result.name;
      },
      error => {
        this.defaultServiceErrorHandling(error);
      }
    );
  }

  /**
   * Uploads the file to the backend with the specified file name
   * @param file
   * @param newFileName
   */
  private uploadFile(file: File, newFileName: string){
    this.horseService.postFile(file, newFileName).subscribe(
      () => {
        console.log("Image successfully uploaded");
      }, error => {
        this.defaultServiceErrorHandling(error)
      });
  }

  /**
   * Generates a random hex with the specified length
   * Inspired from:
   * https://stackoverflow.com/a/27747377
   * @param len
   */
  private generateHex(len: number): string {
    var arr = new Uint8Array((len) / 2)
    window.crypto.getRandomValues(arr)
    return Array.from(arr, this.dec2hex).join('')
  }

  /**
   * Converts a decimal to its hex value
   * @param dec
   */
  private dec2hex(dec: number): string {
    return ('0' + dec.toString(16)).substr(-2)
  }

  private defaultServiceErrorHandling(error: any) {
    console.log(error);
    this.error = true;
    if (error.status === 0) {
      // If status is 0, the backend is probably down
      this.errorMessage = 'The backend seems not to be reachable';
    } else if (error.error.message === 'No message available') {
      // If no detailed error message is provided, fall back to the simple error name
      this.errorMessage = error.error.error;
    } else {
      this.errorMessage = error.error.message;
    }
  }
}
