import { Injectable } from '@angular/core';
import { Horse } from '../dto/horse';
import { HttpClient, HttpParams } from '@angular/common/http';
import { Globals } from '../global/globals';
import { Observable } from 'rxjs';

@Injectable({
  providedIn: 'root'
})
export class HorseService {
  private messageBaseUri: string = this.globals.backendUri + '/horses';

  constructor(private httpClient: HttpClient, private globals: Globals) {
  }

  /**
   * Loads specific horse from the backend
   * @param id of horse to load
   */
  getHorseById(id: number): Observable<Horse> {
    console.log('Load horse details for ' + id);
    return this.httpClient.get<Horse>(this.messageBaseUri + '/' + id);
  }

  /**
   * Loads all horses from the backend
   */
  getAllHorses(): Observable<Array<Horse>> {
    console.log('Load all horses');
    return this.httpClient.get<Array<Horse>>(this.messageBaseUri);
  }

  /**
   * Load all horses from the backend that correspond to a filter
   * @param filter to use
   */
  getFilteredHorses(filter: HttpParams): Observable<Array<Horse>> {
    console.log('Load all filtered horses');
    return this.httpClient.get<Array<Horse>>(this.messageBaseUri, { params: filter });
  }

  /**
   * Adds a specific horse to the backend and loads it if successful
   * @param horse
   */
  addHorse(horse: Horse): Observable<Horse> {
    console.log('Add new horse ' + JSON.stringify(horse));
    return this.httpClient.post<Horse>(this.messageBaseUri, horse);
  }

  /**
   * Update a specific horse with the supplied values
   * @param horse
   */
  updateHorse(horse: Horse): Observable<Horse> {
    console.log('Update horse with id ' + horse.id + ': '+ JSON.stringify(horse));
    return this.httpClient.put<Horse>(this.messageBaseUri + '/' + horse.id, horse);
  }

  /**
   * Delete a horse from the backend
   * @param id
   */
  deleteHorse(id: number): Observable<Object> {
    console.log('Delete horse with id ' + id);
    return this.httpClient.delete(this.messageBaseUri + '/' + id);
  }

  /**
   * Uploads a file to the remote server
   * @param fileToUpload
   * @param newFileName
   */
  postFile(fileToUpload: File, newFileName: string): Observable<Object> {
    // Prepare the form
    var formData: FormData = new FormData();
    formData.append('file', fileToUpload, newFileName);

    // Upload the image and return the result
    return this.httpClient
      .post(this.messageBaseUri + '/upload', formData);
  }
}
